#include <GL/glew.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <vector>

#include <gmath/gmath.h>

#include "gfxapi.h"
#include "global.h"

/* TODO: fix those */
#include "camera.h"
#include "mesh.h"
#include "object.h"
#include "renderer.h"
#include "scene.h"
#include "texture.h"

#include "opengl/opengl.h"
#include "vulkan/vk.h"

/* static functions */

static bool init(Gfx_API api);
static void cleanup();
static void display();

/* glfw callbacks */

static void clbk_key(GLFWwindow *win, int key, int scancode, int action, int mods);
static void clbk_motion(GLFWwindow *win, double x, double y);
static void clbk_mouse(GLFWwindow *win, int button, int action, int mods);
static void clbk_reshape(GLFWwindow *win, int width, int height);

/* global variables */

Mat4 mprojection;

GLFWwindow *win;
int win_w = 800;
int win_h = 600;

float phi = 25;
float theta = 0;
float dist = 4;

ShaderManager *sdr_man;

/* variables */

static float aspect;
static OrbitCamera *camera;

static Scene *scene_cow;
static Renderer *rcow;

static Scene *scene_ground;
static Renderer *rground; // default renderer
static Texture *gskybox;

/* *** */

int main(int argc, char **argv)
{
	Gfx_API api;

	for(int i=0; i<argc; ++i) {
		if(strcmp(argv[i], "-opengl") == 0) {
			api = GFX_GL;
			printf("Backend: OpenGL.\n");
		}
		else if(strcmp(argv[i], "-vulkan") == 0) {
			api = GFX_VK;
			printf("Backend: Vulkan.\n");
		}
		else {
			api = GFX_GL;
			printf("No backend specified. Using OpenGL.\n");
		}
	}

	if(!init(api)) {
		fprintf(stderr, "Failed to initialize program.\n");
		return 1;
	}

	glfwSetKeyCallback(win, clbk_key);
	glfwSetCursorPosCallback(win, clbk_motion);
	glfwSetMouseButtonCallback(win, clbk_mouse);
	glfwSetWindowSizeCallback(win, clbk_reshape);

	glfwGetWindowSize(win, &win_w, &win_h);
	clbk_reshape(win, win_w, win_h);

	while(!glfwWindowShouldClose(win)) {
		display();

		glfwSwapBuffers(win);
		glfwPollEvents();
	}

	cleanup();
	// atexit(cleanup);
	return 0;
}

static bool init(Gfx_API api)
{
	if(!gfx_init(api))
		return false;

	sdr_man = new ShaderManager;

	camera = new OrbitCamera;
	camera->set_orbit_params(phi, theta, dist);

	scene_ground = new Scene;
	if(!scene_ground->load("data/ground.obj")) {
		fprintf(stderr, "Failed to load scene: ground.obj.\n");
		return false;
	}

	scene_cow = new Scene;
	if(!scene_cow->load("data/spot/spot.obj")) {
		fprintf(stderr, "Failed to load scene: spot.obj.\n");
		return false;
	}

	rground = new Renderer;
	rground->camera = camera;
	rground->scene = scene_ground;

	if(!rground->create()) {
		fprintf(stderr, "Failed to create default renderer.\n");
		return false;
	}

	gskybox = gfx_create_texture();
	gskybox->load("data/cubemap/cubemap.jpg");
	rground->set_sky_tex(gskybox);

	rcow = new Renderer;
	rcow->camera = camera;
	rcow->scene = scene_cow;

	if(!rcow->create()) {
		fprintf(stderr, "Failed to create renderer for cows.\n");
		return false;
	}

// TODO delete: debugging
	for(size_t i=0; i<scene_ground->objects.size(); ++i) {
		printf("object: %d\n", (int)i);
		printf("mesh: %s\n", scene_ground->objects[i]->mesh->name.c_str());
		printf("material: %s\n", scene_ground->objects[i]->material->name.c_str());
		printf("transform:\n");
		scene_ground->objects[i]->transform.print();
	}
	return true;
}

static void cleanup()
{
	delete sdr_man;
	delete camera;

	delete scene_cow;
	delete rcow;

	delete scene_ground;
	delete rground;

	gfx_cleanup();
}

static void clbk_key(GLFWwindow *win, int key, int scancode, int action, int mods)
{
	if(key == GLFW_KEY_ESCAPE && action == GLFW_PRESS) {
		glfwSetWindowShouldClose(win, GLFW_TRUE);
	}
}

static double prev_x, prev_y;
static bool button[8];

static void clbk_motion(GLFWwindow *win, double x, double y)
{
	double dx = x - prev_x;
	double dy = y - prev_y;

	prev_x = x;
	prev_y = y;

	if(button[0]) {
		theta += dx;
		phi += dy;

		if(phi < -90)
			phi = -90;
		if(phi > 90)
			phi = 90;
	}

	if(button[1]) {
		dist += dy * 0.1;
		if(dist < 0.0) {
			dist = 0.0;
		}
	}
}

static void clbk_mouse(GLFWwindow *win, int bn, int action, int mods)
{
	button[bn] = action == GLFW_PRESS;
	glfwGetCursorPos(win, &prev_x, &prev_y);
}

static void clbk_reshape(GLFWwindow *win, int width, int height)
{
	gfx_viewport(0, 0, width, height);
	aspect = (float)width / (float)height;
	mprojection = calc_projection_matrix(45, aspect, 0.5, 1000.0);

	win_h = height;
	win_w = width;
}

static void display()
{
	camera->set_orbit_params(phi, theta, dist);

	gfx_clear(0.76, 0.3, 0.43);

	rground->draw();
	rcow->draw();
}